"""
AST to Graphviz Conversion Module.

This module provides utilities for converting an Abstract Syntax Tree (AST),
represented as a nested Python dictionary, to a Graphviz dot graph. The graph
can be displayed inline in a Jupyter notebook.
"""

from typing import Optional, cast

from graphviz import Digraph
from IPython.display import Image, display  # type: ignore[attr-defined]

from astx.base import ReprStruct


def traverse_ast(
    ast: ReprStruct,
    graph: Optional[Digraph] = None,
    parent: Optional[str] = None,
    shape: str = "box",
) -> Digraph:
    """
    Recursively traverse the AST and build a Graphviz graph.

    Parameters
    ----------
    ast : ReprStruct
        The AST as a nested dictionary (full structure version).
    graph : Digraph
        The Graphviz graph object.
    parent : str, optional
        The identifier of the parent node in the graph, by default
        it is an empty string
    shape: str, options: ellipse, box, circle, diamond
        The shape used for the nodes in the graph. Default "box".
    """
    if not graph:
        graph = Digraph()
        graph.attr(rankdir="TB")

    if not isinstance(ast, dict):
        return graph.unflatten(stagger=3)

    for key, full_value in ast.items():
        if not isinstance(full_value, dict):
            continue

        value = full_value.get("value", "")
        metadata = full_value.get("metadata", {})
        ref = ""

        if isinstance(metadata, dict):
            ref = cast(str, metadata.get("ref", ""))

        node_name = f"{hash(key)}_{hash(str(ref))}_{hash(str(value))}"
        graph.node(node_name, key, shape=shape)

        if parent:
            graph.edge(parent, node_name)

        if isinstance(value, dict):
            traverse_ast(value, graph, node_name, shape=shape)
            continue
        elif not isinstance(value, list):
            continue

        for item in value:
            if isinstance(item, dict):
                traverse_ast(item, graph, node_name, shape=shape)
    return graph.unflatten(stagger=3)


def visualize(ast: ReprStruct, shape: str = "box") -> None:
    """
    Visualize the AST using graphviz.

    Parameters
    ----------
    ast: ReprStruct
    shape: str, options: ellipse, box, circle, diamond
        The shape used for the nodes in the graph. Default "box".
    """
    graph = traverse_ast(ast, shape=shape)
    image = Image(graph.pipe(format="png"))  # type: ignore[no-untyped-call]
    display(image)  # type: ignore[no-untyped-call]



"""Module for controle flow AST."""

from typing import Optional

from public import public

from astx.base import (
    NO_SOURCE_LOCATION,
    ASTKind,
    ASTNodes,
    Expr,
    ReprStruct,
    SourceLocation,
    StatementType,
)
from astx.blocks import Block
from astx.variables import InlineVariableDeclaration


@public
class If(StatementType):
    """AST class for `if` statement."""

    condition: Expr
    then: Block
    else_: Optional[Block]

    def __init__(
        self,
        condition: Expr,
        then: Block,
        else_: Optional[Block] = None,
        loc: SourceLocation = NO_SOURCE_LOCATION,
        parent: Optional[ASTNodes] = None,
    ) -> None:
        """Initialize the If instance."""
        super().__init__(loc=loc, parent=parent)
        self.loc = loc
        self.condition = condition
        self.then = then
        self.else_ = else_
        self.kind = ASTKind.IfKind

    def __str__(self) -> str:
        """Return a string representation of the object."""
        return f"If[{self.condition}]"

    def get_struct(self, simplified: bool = False) -> ReprStruct:
        """Return the AST structure of the object."""
        if_condition = self.condition.get_struct(simplified)
        if_then = self.then.get_struct(simplified)

        if self.else_:
            if_else = self.else_.get_struct(simplified)
            if_else_struct = self._prepare_struct("ELSE", if_else, simplified)
        else:
            if_else_struct = {}

        if_condition_struct = self._prepare_struct(
            "CONDITION", if_condition, simplified
        )
        if_then_struct = self._prepare_struct("THEN", if_then, simplified)

        if not isinstance(if_condition_struct, dict):
            raise Exception("`if_condition` struct is not a valid object.")

        if not isinstance(if_then_struct, dict):
            raise Exception("`if_then` struct is not a valid object.")

        if not isinstance(if_else_struct, dict):
            raise Exception("`if_else` struct is not a valid object.")

        key = "IF-STMT"
        value: ReprStruct = {
            **if_condition_struct,
            **if_then_struct,
            **if_else_struct,
        }

        return self._prepare_struct(key, value, simplified)


@public
class ForRangeLoop(StatementType):
    """AST class for `For` Loop Range statement."""

    variable: InlineVariableDeclaration
    start: Expr
    end: Expr
    step: Expr
    body: Block

    def __init__(
        self,
        variable: InlineVariableDeclaration,
        start: Expr,
        end: Expr,
        step: Expr,
        body: Block,
        loc: SourceLocation = NO_SOURCE_LOCATION,
        parent: Optional[ASTNodes] = None,
    ) -> None:
        """Initialize the ForStmt instance."""
        super().__init__(loc=loc, parent=parent)
        self.variable = variable
        self.start = start
        self.end = end
        self.step = step
        self.body = body
        self.kind = ASTKind.ForKind

    def __str__(self) -> str:
        """Return a string that represents the object."""
        start = self.start
        end = self.end
        step = self.step
        var_name = self.variable.name
        return f"ForRangeLoop({var_name}=[{start}:{end}:{step}])"

    def get_struct(self, simplified: bool = False) -> ReprStruct:
        """Return the AST structure of the object."""
        for_start = self.start.get_struct(simplified)
        for_end = self.end.get_struct(simplified)
        for_step = self.step.get_struct(simplified)
        for_body = self.body.get_struct(simplified)

        for_start_struct = self._prepare_struct("start", for_start, simplified)
        for_end_struct = self._prepare_struct("end", for_end, simplified)
        for_step_struct = self._prepare_struct("step", for_step, simplified)
        for_body_struct = self._prepare_struct("body", for_body, simplified)

        if not isinstance(for_start_struct, dict):
            raise Exception("`for_start` struct is not a valid object.")

        if not isinstance(for_end_struct, dict):
            raise Exception("`for_end` struct is not a valid object.")

        if not isinstance(for_step_struct, dict):
            raise Exception("`for_step` struct is not a valid object.")

        if not isinstance(for_body_struct, dict):
            raise Exception("`for_body` struct is not a valid object.")

        key = "FOR-RANGE-STMT"
        value: ReprStruct = {
            **for_start_struct,
            **for_end_struct,
            **for_step_struct,
            **for_body_struct,
        }

        return self._prepare_struct(key, value, simplified)


@public
class ForCountLoop(StatementType):
    """
    AST class for a simple Count-Controlled `For` Loop statement.

    This is a very basic `for` loop, used by languages like C or C++.
    """

    initializer: InlineVariableDeclaration
    condition: Expr
    update: Expr
    body: Block

    def __init__(
        self,
        initializer: InlineVariableDeclaration,
        condition: Expr,
        update: Expr,
        body: Block,
        loc: SourceLocation = NO_SOURCE_LOCATION,
        parent: Optional[ASTNodes] = None,
    ) -> None:
        """Initialize the ForStmt instance."""
        super().__init__(loc=loc, parent=parent)
        self.initializer = initializer
        self.condition = condition
        self.update = update
        self.body = body
        self.kind = ASTKind.ForKind

    def __str__(self) -> str:
        """Return a string that represents the object."""
        init = self.initializer
        cond = self.condition
        update = self.update
        return f"ForCountLoop({init};{cond};{update})"

    def get_struct(self, simplified: bool = False) -> ReprStruct:
        """Return the AST structure of the object."""
        for_init = self.initializer.get_struct(simplified)
        for_cond = self.condition.get_struct(simplified)
        for_update = self.update.get_struct(simplified)
        for_body = self.body.get_struct(simplified)

        for_init_struct = self._prepare_struct(
            "initializer", for_init, simplified
        )
        for_cond_struct = self._prepare_struct(
            "condition", for_cond, simplified
        )
        for_update_struct = self._prepare_struct(
            "update", for_update, simplified
        )
        for_body_struct = self._prepare_struct("body", for_body, simplified)

        if not isinstance(for_init_struct, dict):
            raise Exception("`for_init` struct is not a valid object.")

        if not isinstance(for_cond_struct, dict):
            raise Exception("`for_cond` struct is not a valid object.")

        if not isinstance(for_update_struct, dict):
            raise Exception("`for_update` struct is not a valid object.")

        if not isinstance(for_body_struct, dict):
            raise Exception("`for_body` struct is not a valid object.")

        key = "FOR-COUNT-STMT"
        value: ReprStruct = {
            **for_init_struct,
            **for_cond_struct,
            **for_update_struct,
            **for_body_struct,
        }

        return self._prepare_struct(key, value, simplified)



"""Define ASTx for more broader scope."""

from __future__ import annotations

import copy

from typing import cast

from public import public

from astx.base import (
    AST,
    NO_SOURCE_LOCATION,
    ASTKind,
    ASTNodes,
    Expr,
    ReprStruct,
    SourceLocation,
)
from astx.blocks import Block


@public
class Target(Expr):
    """Define the Architecture target for the program."""

    datalayout: str
    triple: str

    def __init__(self, datalayout: str, triple: str) -> None:
        """Initialize the AST instance."""
        self.datalayout = datalayout
        self.triple = triple

    def get_struct(self, simplified: bool = False) -> ReprStruct:
        """Return the AST structure of the object."""
        key = "TARGET"
        value = f"{self.datalayout}, {self.triple}"
        return self._prepare_struct(key, value, simplified)


@public
class Module(Block):
    """AST main expression class."""

    name: str

    def __init__(
        self,
        name: str = "main",
        loc: SourceLocation = NO_SOURCE_LOCATION,
    ) -> None:
        """Initialize the AST instance."""
        super().__init__(name=name, loc=loc)
        self.kind = ASTKind.ModuleKind

    def __str__(self) -> str:
        """Return the string representation of the object."""
        return f"Module[{self.name}]"

    @property
    def block(self) -> list[AST]:
        """Define an alias for self.nodes."""
        return self.nodes

    def get_struct(self, simplified: bool = False) -> ReprStruct:
        """Return the AST structure of the object."""
        block_node = []

        for node in self.nodes:
            block_node.append(node.get_struct(simplified))

        key = f"MODULE[{self.name}]"
        value = cast(ReprStruct, block_node)

        return self._prepare_struct(key, value, simplified)


@public
class Package(ASTNodes):
    """AST class for Package."""

    name: str
    modules: list[Module]
    packages: list[Package]

    def __init__(
        self,
        name: str = "main",
        modules: list[Module] = [],
        packages: list[Package] = [],
        loc: SourceLocation = NO_SOURCE_LOCATION,
    ) -> None:
        """Initialize the AST instance."""
        super().__init__(loc=loc)
        self.name = name
        self.modules = copy.deepcopy(modules)
        self.packages = copy.deepcopy(packages)

    def __str__(self) -> str:
        """Return the string representation of the object."""
        return f"PACKAGE[{self.name}]"

    def get_struct(self, simplified: bool = False) -> ReprStruct:
        """Return the AST structure of the object."""
        packages = []
        modules = []

        for package in self.packages:
            packages.append(package.get_struct(simplified))

        for module in self.modules:
            modules.append(module.get_struct(simplified))

        key = str(self)
        value = cast(
            ReprStruct,
            {
                "modules": modules,
                "packages": packages,
            },
        )

        return self._prepare_struct(key, value, simplified)


@public
class Program(Package):
    """AST class for Program."""

    target: Target

    def __init__(
        self,
        name: str = "main",
        target: Target = Target("", ""),
        modules: list[Module] = [],
        packages: list[Package] = [],
        loc: SourceLocation = NO_SOURCE_LOCATION,
    ) -> None:
        """Initialize the AST instance."""
        super().__init__(
            name=name, modules=modules, packages=packages, loc=loc
        )
        self.target = copy.deepcopy(target)

    def __str__(self) -> str:
        """Return the string representation of the object."""
        return f"PROGRAM[{self.name}]"
